package holeg.ui.controller;

import holeg.model.*;
import holeg.ui.model.GuiSettings;
import holeg.model.Model;
import holeg.utility.math.vector.Vec2i;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * Controller for the Canvas.
 *
 * @author Gruppe14
 */
public class CanvasController {

    private final Control control;

    /**
     * Constructor.
     *
     * @param control the Controller
     *
     */
    public CanvasController(Control control) {
        this.control = control;
    }

    /**
     * Add an CpsObject to the model and notify the ObjectListener for update.
     *
     * @param object CpsObject to be added.
     */
    public void addObject(GroupNode groupNode, AbstractCanvasObject object) {
        groupNode.add(object);
    }


    /**
     * Deletes an CpsObject on the Canvas and its connections.
     *
     * @param obj AbstractCpsObject
     */
    public void deleteObject(AbstractCanvasObject obj) {
        if(obj instanceof GroupNode groupNode) {
            deleteAllObjectsInGroupNode(groupNode);
        }else{
            removeAllConnectionsFromObject(obj);
        }
        obj.getGroupNode().ifPresent(groupNode -> groupNode.remove(obj));
    }

    public void deleteObjects(Collection<AbstractCanvasObject> objects) {
		objects.forEach(this::deleteObject);
	}

    /**
     * Replaces {@code toBeReplaced} by {@code by} on the canvas
     *
     * @param toBeReplaced the object that will be replaced
     * @param by           the object that will replace it
     */
    public void replaceObject(AbstractCanvasObject toBeReplaced, AbstractCanvasObject by) {
        //Replace edges
        for (Edge edge : control.getModel().getEdgesOnCanvas()) {
            if (edge.getA() == toBeReplaced && edge.getB() != by) {
                edge.setA(by);
            } else if (edge.getB() == toBeReplaced && edge.getA() != by) {
                edge.setB(by);
            }
        }
        // set Position of by to exactly toBeReplaced
        by.setPosition(toBeReplaced.getPosition());
        deleteObject(toBeReplaced);
    }

    /**
     * Add an edge to the Canvas.
     *
     * @param edge the edge
     */
    public void addEdgeOnCanvas(Edge edge) {
        control.getModel().getEdgesOnCanvas().add(edge);
    }

    /**
     * Removes an Edge from the Canvas.
     *
     * @param edge the edge to remove
     */
    public void removeEdgesOnCanvas(Edge edge) {
        control.getModel().getEdgesOnCanvas().remove(edge);
    }

    /**
     * Paste all Selected Objects.
     */
    public void pasteObjects(Vec2i p) {
        //TODO(Tom2022-01-27): paste
        GuiSettings.getSelectedObjects().clear();
    }

    /**
     * Cut all Selected Objects.
     */
    public void cutObjects() {
        GuiSettings.setClipboardObjects(new HashSet<>(GuiSettings.getSelectedObjects()));
        for (AbstractCanvasObject cps : GuiSettings.getClipboardObjects()) {
            deleteObject(cps);
        }
        GuiSettings.getSelectedObjects().clear();
    }


    /**
     * Some cleaning Algorithm which traverses the UpperNode through BFS Can be
     * extended with other cleaning stuff No need for coloring since there tree
     * is only directed in one direction
     */
    public void deleteAllObjectsInGroupNode(GroupNode node) {
        Set<AbstractCanvasObject> objectsInGroupNode = node.getAllObjectsRecursive().collect(Collectors.toSet());
        control.getModel().getEdgesOnCanvas().removeIf(edge -> objectsInGroupNode.contains(edge.getA()) || objectsInGroupNode.contains(edge.getB()));
    }

    public void removeAllConnectionsFromObject(AbstractCanvasObject obj) {
        control.getModel().getEdgesOnCanvas().removeIf(edge -> edge.getA() == obj || edge.getB() == obj);
    }

    public void addGroupNode(GroupNode groupNode, Set<AbstractCanvasObject> toGroup) {
        //TODO(Tom2022-01-27):
    }
}
